from abc import ABC, abstractmethod
from typing import Dict, Any, Union, List, Tuple, Optional, Callable, Iterable, Iterator, Literal
from typing_extensions import Self
from pathlib import Path
from sep import Background
import numpy as np
from astropy.table import QTable
from astropy.nddata import CCDData
from astropy.coordinates import SkyCoord

from myraflib.data_model import Data
from myraflib.utils import NUMERIC, NUMERICS, HEADER_ANY


class DataArray(ABC):
    data: Iterable[Data]

    @abstractmethod
    def __str__(self) -> str:
        """String representation of Data"""

    @abstractmethod
    def __repr__(self) -> str:
        """Representation of Data"""

    # @abstractmethod
    # def __del__(self) -> None:
    #     """Destructor of data"""
    #
    @abstractmethod
    def __abs__(self) -> Self:
        """Absolute value of data"""

    @abstractmethod
    def __add__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (+)"""

    @abstractmethod
    def __radd__(self, other: Union[Self, NUMERIC]) -> Self:
        """Right operator overloading (+)"""

    @abstractmethod
    def __sub__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (-)"""

    @abstractmethod
    def __rsub__(self, other: Union[Self, NUMERIC]) -> Self:
        """Right operator overloading (-)"""

    @abstractmethod
    def __mul__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (*)"""

    @abstractmethod
    def __rmul__(self, other: Union[Self, NUMERIC]) -> Self:
        """Right operator overloading (*)"""

    @abstractmethod
    def __truediv__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (/)"""

    @abstractmethod
    def __pow__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        """Operator overloading (**)"""

    #
    @abstractmethod
    def __mod__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        """Operator overloading (%)"""

    @abstractmethod
    def __rmod__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        """Right operator overloading (%)"""

    @abstractmethod
    def __neg__(self) -> Self:
        """Negative operator overloading"""

    @abstractmethod
    def __pos__(self) -> Self:
        """Positive operator overloading"""

    @abstractmethod
    def __contains__(self, item: Data) -> bool:
        """Checks if item is in DataArray"""

    @abstractmethod
    def __setitem__(self, key: Union[int, slice], value: Data) -> None:
        """Sets value to the given index"""

    @abstractmethod
    def __delitem__(self, key: Union[int, slice]) -> None:
        """Deletes value to the given index"""

    @classmethod
    @abstractmethod
    def from_paths(cls, paths: List[str]) -> Self:
        """Returns a DataArray from a list of paths"""

    @classmethod
    @abstractmethod
    def from_pattern(cls, pattern: str) -> Self:
        """Returns a DataArray from a pattern (glob)"""

    @classmethod
    @abstractmethod
    def sample(cls, n=10) -> Self:
        """Creates a sample DataArray"""

    @abstractmethod
    def merge(self, other: Self) -> Self:
        """Returns merged DataArrays"""

    @abstractmethod
    def append(self, other: Data) -> Self:
        """Returns an appended DataArray"""

    @property
    @abstractmethod
    def header(self) -> QTable:
        """Returns headers of DataArray"""

    @property
    @abstractmethod
    def data(self) -> Iterator[np.ndarray]:
        """Returns data of DataArray"""

    @property
    @abstractmethod
    def ccd(self) -> Iterator[CCDData]:
        """Returns CCD data of DataArray"""

    @abstractmethod
    def value(self, x: int, y: int) -> np.ndarray:
        """Returns value of a given coordinates"""

    @property
    @abstractmethod
    def files(self) -> List[str]:
        """Returns the fits files' paths"""

    @property
    @abstractmethod
    def stats(self) -> QTable:
        """Returns a DataFrame of statistics of DataArray"""

    @abstractmethod
    def background(self) -> Iterator[Background]:
        """Returns the background of the given fits file"""

    @abstractmethod
    def cosmic_cleaner(self, sigclip: float = 4.5, sigfrac: float = 0.3, objlim: float = 5.0, gain: float = 1.0,
                       readnoise: float = 6.5, satlevel: float = 65535.0, pssl: float = 0.0, niter: int = 4,
                       sepmed: bool = True, cleantype: Literal["median", "medmask", "meanmask", "idw"] = 'meanmask',
                       fsmode: Literal["median", "convolve"] = 'median',
                       psfmodel: Literal["gauss", "moffat", "gaussx", "gaussy"] = 'gauss',
                       psffwhm: float = 2.5, psfsize: int = 7, psfk: np.ndarray = None, psfbeta: float = 4.765,
                       gain_apply: bool = True) -> Self:
        """Cleans the fits file from cosmic rays"""

    @abstractmethod
    def hedit(self, keys: Union[str, List[str]], values: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None,
              delete: bool = False, value_is_key: bool = False) -> Self:
        """Edits header of given files"""

    @abstractmethod
    def hselect(self, fields: Union[str, List[str]]) -> QTable:
        """Returns a Table of selected headers of DataArray"""

    @abstractmethod
    def save(self, path: Union[str, Path], overwrite: bool = False) -> Self:
        """Saves DataArray to given path"""

    @abstractmethod
    def abs(self) -> Self:
        """Return absolute value of DataArray"""

    @abstractmethod
    def add(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]) -> Self:
        """Adds either another numeric, Data, DataArray or list of numerics to the given DataArray"""

    @abstractmethod
    def sub(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]) -> Self:
        """Subtracts either another numeric, Data, DataArray or list of numerics from the given DataArray"""

    @abstractmethod
    def mul(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]) -> Self:
        """Multiplies either another numeric, Data, DataArray or list of numerics to the given DataArray"""

    @abstractmethod
    def div(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]) -> Self:
        """Divides either another numeric, Data, DataArray or list of numerics to the given DataArray"""

    @abstractmethod
    def pow(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]) -> Self:
        """Raises the data to given value"""

    def mod(self, other: Union[Self, Data, NUMERIC, List[Union[Data, NUMERIC]]]):
        """Modular arithmetic operation"""

    @abstractmethod
    def imarith(self, other: Union[Self, NUMERIC], operator: Literal["+", "-", "*", "/", "**", "^", "%"]) -> Self:
        """Makes arithmetic operation to fits"""

    @abstractmethod
    def extract(self, index: int = 0, detection_sigma: float = 5.0, min_area: float = 5.0) -> QTable:
        """Extracts sources from fits"""

    @abstractmethod
    def daofind(self, index: int = 0, sigma: float = 3.0, fwhm: float = 3.0, threshold: float = 5.0) -> QTable:
        """Extracts sources from fits using daofind"""

    @abstractmethod
    def align(self, reference: Union[Data, int] = 0, max_control_points: int = 50, min_area: int = 5) -> Self:
        """Aligns the DataArray by another Data"""

    @abstractmethod
    def shift(self, xs: Union[List[int], int], ys: Union[List[int], int]) -> Self:
        """Shifts the DataArray by the given numerics"""

    @abstractmethod
    def rotate(self, angle: Union[List[NUMERIC], NUMERIC]) -> Self:
        """Rotates the DataArray by given angle"""

    @abstractmethod
    def crop(self, x: Union[int, List[int]], y: Union[int, List[int]],
             width: Union[int, List[int]], height: Union[int, List[int]]) -> Self:
        """Crops the DataArray by given coordinates"""

    @abstractmethod
    def bin(self, binning_factor: Union[Union[int, Tuple[int, int]], List[Union[int, Tuple[int, int]]]],
            func: Callable[..., float] = np.mean) -> Self:
        """Bin the DataArray"""

    @abstractmethod
    def show(self, points: QTable = None, scale: bool = True, interval: float = 1.0) -> None:
        """Shows DataArray"""

    @abstractmethod
    def solve_field(self, api_key: str, reference: Union[Data, int] = 0, solve_timeout: int = 120,
                    force_image_upload: bool = False, max_control_points: int = 50, min_area: int = 5,
                    soft_solve: bool = True) -> Self:
        """WCS solves"""

    @abstractmethod
    def group_by(self, groups: Union[str, List[str]]) -> Dict[Any, Self]:
        """Groups the DataArray"""

    @abstractmethod
    def combine(self, method: Literal["average", "mean", "median", "sum"] = "average",
                clipping: Literal[None, "sigma", "minmax"] = None,
                weights: Optional[List[Union[float, int]]] = None) -> Data:
        """Combines the DataArray to Data"""

    @abstractmethod
    def zero_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None) -> Data:
        """Combines the DataArray to Data (zerocombine)"""

    @abstractmethod
    def dark_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None,
                     weights: Optional[Union[List[str], List[Union[float, int]]]] = None) -> Data:
        """Combines the DataArray to Data (darkcombine)"""

    @abstractmethod
    def flat_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None,
                     weights: Optional[Union[List[str], List[Union[float, int]]]] = None) -> Data:
        """Combines the DataArray to Data (flatcombine)"""

    @abstractmethod
    def zero_correction(self, master_zero: Self, force: bool = False) -> Self:
        """Does zero correction of the data"""

    @abstractmethod
    def dark_correction(self, master_dark: Self, exposure: Optional[str] = None, force: bool = False) -> Self:
        """Dark corrects the fits"""

    @abstractmethod
    def flat_correction(self, master_flat: Self, exposure: Optional[str] = None, force: bool = False) -> Self:
        """Flat corrects the fits"""

    @abstractmethod
    def photometry_sep(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None
                       ) -> QTable:
        """Does a photometry using sep"""

    @abstractmethod
    def photometry_phu(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None
                       ) -> QTable:
        """Does a photometry using photutils"""

    @abstractmethod
    def photometry(self, xs: NUMERICS, ys: NUMERICS, rs: NUMERICS,
                   headers: Optional[Union[str, List[str]]] = None,
                   exposure: Optional[Union[str, float, int]] = None
                   ) -> QTable:
        """Does a photometry using both sep and photutils"""

    @abstractmethod
    def pixels_to_skys(self, xs: Union[List[Union[int, float]], int, float],
                       ys: Union[List[Union[int, float]], int, float]) -> QTable:
        """Calculate Sky Coordinate of given Pixel"""

    @abstractmethod
    def skys_to_pixels(self, skys: SkyCoord) -> QTable:
        """Calculate Pixel Coordinate of given Sky"""



import itertools
import math
from copy import copy
from logging import Logger, getLogger
from pathlib import Path

from typing import Optional, Dict, Any, Union, List, Literal, Tuple, Callable
from typing_extensions import Self

import numpy as np
from scipy import ndimage

from matplotlib import pyplot as plt
from mpl_point_clicker import clicker

import astroalign

from astropy.coordinates import SkyCoord
from astropy import units
from astropy.units import Quantity
from astropy.visualization import ZScaleInterval
from astropy.stats import sigma_clipped_stats
from astropy.table import QTable, vstack
from astropy.wcs import WCS
from astropy.wcs.utils import fit_wcs_from_points
from astropy.io import fits as fts
from astropy.io.fits import Header
from astropy.nddata import CCDData, block_reduce

from sep import extract as sep_extract, Background, sum_circle

from ccdproc import cosmicray_lacosmic, subtract_bias, subtract_dark, flat_correct

from photutils.detection import DAOStarFinder
from photutils.aperture import CircularAperture, aperture_photometry
from photutils.utils import calc_total_error

from myraflib.data_model import Data
from myraflib.error import NothingToDo, OperatorError, NumberOfElementError, Unsolvable, AlignError, OverCorrection, \
    CardNotFound
from myraflib.utils import Fixer, NUMERIC, HEADER_ANY, NUMERICS


class Fits(Data):
    ZMag = 25

    def __init__(self, path: Path, logger: Optional[Logger] = None) -> None:

        if logger is None:
            self.logger = getLogger(__file__)
        else:
            self.logger = logger

        if not path.exists():
            self.logger.error("File does not exist")
            raise FileNotFoundError("File does not exist")

        self.is_temp = False
        self.path = path
        self.hdu = fts.open(self.path.absolute(), mode="update")

        if self.data.ndim != 2:
            self.logger.error("Data must be 2D")
            raise ValueError("Data must be 2D")

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.path})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.from_path({self.path.absolute()})"

    def __del__(self):
        try:
            if self.is_temp:
                if self.path.exists():
                    self.path.unlink()
        except Exception as e:
            self.logger.warning(e)

    def __abs__(self) -> Self:
        return self.abs()

    def __add__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.add(other)

    def __iadd__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.add(other)

    def __sub__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.sub(other)

    def __isub__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.sub(other)

    def __mul__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.mul(other)

    def __imul__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.mul(other)

    def __truediv__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.div(other)

    def __idiv__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.div(other)

    def __pow__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        return self.pow(power)

    def __ipow__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.pow(other)

    def __mod__(self, other: Union[Self, NUMERIC], modulo=None) -> Self:
        return self.mod(other)

    def __imod__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.mod(other)

    def __eq__(self, other: Self) -> bool:
        return self.is_same(other)

    def __ne__(self, other: Self) -> bool:
        return not self.is_same(other)

    def __neg__(self) -> Self:
        return self.mul(-1)

    def __pos__(self) -> Self:
        return self

    def flux_to_mag(self, flux: Union[int, float],
                    flux_error: Union[int, float],
                    exptime: Union[int, float]
                    ) -> Tuple[Union[int, float], Union[int, float]]:
        self.logger.info("Converting flux to mag")
        mag = -2.5 * np.log10(flux)
        if exptime != 0:
            mag += 2.5 * np.log10(exptime)

        if flux_error <= 0:
            mag_err = 0.0
        else:
            mag_err = 1.0857 * flux_error / flux

        if np.isinf(mag_err):
            mag_err = 0

        return mag + self.ZMag, mag_err

    @property
    def file(self) -> str:
        self.logger.info("Getting file path")
        return self.path.absolute().__str__()

    @classmethod
    def from_data_header(cls, data: np.ndarray, header: Optional[Header] = None,
                         output: Optional[str] = None, override: bool = False) -> Self:
        new_output = Fixer.output(output=output, override=override)
        fts.writeto(new_output, data, header=header, output_verify="silentfix")
        fits = cls.from_path(new_output)

        fits.is_temp = output is None
        return fits

    @classmethod
    def from_path(cls, path: str) -> Self:
        return cls(Path(path))

    @classmethod
    def sample(cls) -> Self:
        file = str(Path(__file__).parent / "sample.fits")
        data = fts.getdata(file)
        header = fts.getheader(file)
        return cls.from_data_header(data, header=header)

    def is_same(self, other: Self) -> bool:
        self.logger.info("Checking if two Fits are the same")
        return np.array_equal(
            self.data, other.data
        )

    @property
    def ccd(self) -> CCDData:
        self.logger.info("Creating CCDData of the fits file")
        return CCDData(self.data, unit="adu")

    @property
    def data(self) -> np.ndarray:
        self.logger.info("Getting data")
        return self.hdu[0].data

    def value(self, x: int, y: int) -> NUMERIC:
        self.logger.info("Getting value from fits file")
        value = float(self.data[x][y])

        if value.is_integer():
            return int(value)

        return value

    @property
    def header(self) -> Dict[str, Any]:
        self.logger.info("Getting header")
        header = self.hdu[0].header
        return {
            card: header[card]
            for card in header
            if card not in ['COMMENT', 'HISTORY']
        }

    @property
    def stats(self) -> Dict[str, Any]:
        self.logger.info("Getting stats")
        data = self.data
        w, h = data.shape
        return {
            "size": data.size,
            "width": w,
            "height": h,
            "min": data.min(),
            "mean": data.mean(),
            "std": data.std(),
            "max": data.max(),
        }

    def background(self) -> Background:
        self.logger.info("Getting background")
        return Background(self.data.astype(float))

    def cosmic_cleaner(self, sigclip: float = 4.5, sigfrac: float = 0.3, objlim: float = 5.0, gain: float = 1.0,
                       readnoise: float = 6.5, satlevel: float = 65535.0, pssl: float = 0.0, niter: int = 4,
                       sepmed: bool = True, cleantype: Literal["median", "medmask", "meanmask", "idw"] = 'meanmask',
                       fsmode: Literal["median", "convolve"] = 'median',
                       psfmodel: Literal["gauss", "moffat", "gaussx", "gaussy"] = 'gauss',
                       psffwhm: float = 2.5, psfsize: int = 7, psfk: np.ndarray = None, psfbeta: float = 4.765,
                       gain_apply: bool = True, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Cleaning data")

        cleaned_data, _ = cosmicray_lacosmic(
            self.data, sigclip=sigclip, sigfrac=sigfrac, objlim=objlim, gain=gain, readnoise=readnoise,
            satlevel=satlevel, niter=niter, sepmed=sepmed, cleantype=cleantype, fsmode=fsmode, psfmodel=psfmodel,
            psffwhm=psffwhm, psfsize=psfsize, psfk=psfk, psfbeta=psfbeta, gain_apply=gain_apply
        )

        if output is None:
            self.hdu[0].data = cleaned_data.astype(int).value
            self.hdu.flush()
            return self

        return self.from_data_header(cleaned_data.value, header=self.hdu[0].header, output=output, override=override)

    def hedit(self, keys: Union[str, List[str]], values: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None,
              comments: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None, delete: bool = False,
              value_is_key: bool = False, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Editing the header")

        header = copy(self.hdu[0].header)

        if delete:
            if isinstance(keys, str):
                keys = [keys]

            for key in keys:
                if key in header:
                    del header[key]
                else:
                    self.logger.warning("Key does not exist")

        else:
            if values is None:
                self.logger.error("Delete is False and Value is not given")
                raise NothingToDo("Delete is False and Value is not given")

            keys_to_use, values_to_use, comment_to_use = Fixer.key_value_pair(keys, values, comments)

            if len(keys_to_use) != len(values_to_use):
                self.logger.error("List of keys and values must be equal in length")
                raise ValueError("List of keys and values must be equal in length")

            for key, value, comment in zip(keys_to_use, values_to_use, comment_to_use):
                if value_is_key:
                    header[key] = header[value]
                else:
                    header[key] = value

                header.comments[key] = comment

        if output is None:
            self.hdu[0].header = header
            self.hdu.flush()
            return self

        return self.from_data_header(self.data, header=header, output=output, override=override)

    def save(self, path: Union[str, Path], overwrite: bool = False) -> Self:
        self.logger.info("Saving the fits file")
        if isinstance(path, str):
            new_path = path
        else:
            new_path = path.absolute().__str__()

        return self.from_data_header(self.data, header=self.hdu[0].header, output=new_path, override=overwrite)

    def abs(self, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Getting the absolute value of the fits file")
        new_data = abs(self.data)

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def add(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Adding the value to the fits file")
        if isinstance(other, self.__class__):
            new_data = self.data + other.data
        else:
            new_data = self.data + other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def sub(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Subtracting the value from the fits file")
        if isinstance(other, self.__class__):
            new_data = self.data - other.data
        else:
            new_data = self.data - other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def mul(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Multiplying the value with the fits file")
        if isinstance(other, self.__class__):
            new_data = self.data * other.data
        else:
            new_data = self.data * other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def div(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Dividing the value with the fits file")

        if isinstance(other, self.__class__):
            new_data = self.data / other.data
        else:
            new_data = self.data / other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def pow(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Getting the power of the fits with given value")
        if isinstance(other, self.__class__):
            new_data = self.data ** other.data
        else:
            new_data = self.data ** other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def mod(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False):
        self.logger.info("Getting the mod of the fits with given value")
        if isinstance(other, self.__class__):
            new_data = self.data % other.data
        else:
            new_data = self.data % other

        if output is None:
            self.hdu[0].data = new_data.astype(int)
            self.hdu.flush()
            return self

        return self.from_data_header(new_data.astype(int), header=self.hdu[0].header, output=output, override=override)

    def imarith(self, other: Union[Self, NUMERIC], operator: Literal["+", "-", "*", "/", "**", "^", "%"],
                output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Arithmetic operation")
        if operator == "+":
            return self.add(other, output=output, override=override)
        elif operator == "-":
            return self.sub(other, output=output, override=override)
        elif operator == "*":
            return self.mul(other, output=output, override=override)
        elif operator == "/":
            return self.div(other, output=output, override=override)
        elif operator == "^" or operator == "**":
            return self.pow(other, output=output, override=override)
        elif operator == "%":
            return self.mod(other, output=output, override=override)
        else:
            self.logger.error(f"Unknown operator ({operator})")
            raise OperatorError(f"Unknown operator ({operator})")

    def extract(self, detection_sigma: float = 5.0, min_area: float = 5.0) -> QTable:
        self.logger.info("Extracting sources from fits (sep)")
        bkg = self.background()
        thresh = detection_sigma * bkg.globalrms
        sources = sep_extract(self.data - bkg.back(), thresh, minarea=min_area)
        sources.sort(order="flux")

        if len(sources) < 0:
            self.logger.error("No source was found")
            raise NumberOfElementError("No source was found")

        table = QTable(sources)[["x", "y"]]
        table['x'].unit = units.pix
        table['y'].unit = units.pix
        return table

    def daofind(self, sigma: float = 3.0, fwhm: float = 3.0, threshold: float = 5.0) -> QTable:
        self.logger.info("Extracting sources from fits (daofind)")
        mean, median, std = sigma_clipped_stats(self.data, sigma=sigma)
        daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
        sources = daofind(self.data - median)

        if len(sources) < 0:
            self.logger.error("No source was found")
            raise NumberOfElementError("No source was found")

        table = QTable(sources)[["xcentroid", "ycentroid"]]

        table.rename_column('xcentroid', 'x')
        table.rename_column('ycentroid', 'y')

        table['x'].unit = units.pix
        table['y'].unit = units.pix

        return table

    def align(self, reference: Self, max_control_points: int = 50, detection_sigma: float = 5.0, min_area: int = 5,
              output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Aligning the fits with given fits")

        try:
            data = self.data.astype(float)
            reference_data = reference.data.astype(float)
            w = WCS(self.hdu[0].header)

            t, (source_list, target_list) = astroalign.find_transform(
                source=data,
                target=reference_data,
                max_control_points=max_control_points,
                min_area=min_area,
            )

            registered_image, _ = astroalign.apply_transform(
                t, data, reference_data, 0, False
            )

            try:

                xs = source_list[:, 0]
                ys = source_list[:, 1]

                new_xs = target_list[:, 0]
                new_ys = target_list[:, 1]

                skys = w.pixel_to_world(xs.tolist(), ys.tolist())
                w = fit_wcs_from_points([new_xs, new_ys], skys)
            except Unsolvable:
                self.logger.info("No WCS found in header")
            except AttributeError as e:
                self.logger.info(e)

            temp_header = Header()
            # temp_header.extend(self.pure_header(), unique=True, update=True)
            temp_header.extend(w.to_header(), unique=True, update=True)

            if output is None:
                self.hdu[0].data = registered_image.astype(int)
                self.hdu.header = temp_header
                self.hdu.flush()
                return self

            return self.from_data_header(registered_image.astype(int), header=temp_header,
                                         output=output, override=override)


        except Exception as e:
            print(e)
            self.logger.error("Cannot align two images")
            raise AlignError("Cannot align two images")

    def show(self, ax: plt.Axes = None, sources: QTable = None, scale: bool = True) -> None:
        self.logger.info("Showing fits")

        zscale = ZScaleInterval() if scale else lambda x: x

        if ax is not None:
            ax.imshow(zscale(self.data), cmap="Greys_r")
            return

        plt.imshow(zscale(self.data), cmap="Greys_r")

        if sources is not None:
            plt.scatter(sources["x"].value, sources["y"].value)

        plt.xticks([])
        plt.yticks([])
        plt.show()

    def solve_field(self, api_key: str, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Solving plate")
        if output is None:
            return self

        return self.save(output, overwrite=override)

    def coordinate_picker(self, scale: bool = True) -> QTable:
        self.logger.info("Coordinate picker")
        zscale = ZScaleInterval() if scale else lambda x: x

        fig, ax = plt.subplots(constrained_layout=True)

        ax.imshow(zscale(self.data), cmap="Greys_r")
        klkr = clicker(ax, ["source"], markers=["o"])
        plt.show()

        if len(klkr.get_positions()["source"]) == 0:
            empty_table = QTable()
            empty_table['x'] = Quantity([], unit='pix')  # An empty column with unit meters
            empty_table['y'] = Quantity([], unit='pix')  # An empty column with unit seconds
            return empty_table

        return QTable(
            Quantity(klkr.get_positions()["source"], unit='pix'),
            names=["x", "y"]
        )

    def shift(self, x: NUMERIC, y: NUMERIC, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Shifting fits")
        fill_val = np.median(self.data)
        shifted_data = np.roll(self.data, x, axis=1)
        if x < 0:
            shifted_data[:, x:] = fill_val
        elif x > 0:
            shifted_data[:, 0:x] = fill_val

        shifted_data = np.roll(shifted_data, y, axis=0)
        if y < 0:
            shifted_data[y:, :] = fill_val
        elif y > 0:
            shifted_data[0:y, :] = fill_val

        w = WCS(self.hdu[0].header)

        try:
            highest = min(self.data.shape)
            xs = np.random.randint(0, highest, 20)
            ys = np.random.randint(0, highest, 20)

            skys = w.pixel_to_world(xs.tolist(), ys.tolist())
            w = fit_wcs_from_points([xs + x, ys + y], skys)
        except:
            self.logger.error("No WCS found in header")
            raise Unsolvable("No WCS found in header")

        temp_header = Header()
        # temp_header.extend(self.pure_header(), unique=True, update=True)
        temp_header.extend(w.to_header(), unique=True, update=True)

        if output is None:
            self.hdu[0].data = shifted_data.astype(int)
            self.hdu[0].header = temp_header
            self.hdu.flush()
            return self

        return self.from_data_header(shifted_data.astype(int), header=temp_header, output=output, override=override)

    def rotate(self, angle: NUMERIC, output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Rotating fits")
        fill_val = np.median(self.data)
        angle_degree = angle * 180 / math.pi
        rotated_data = ndimage.rotate(self.data, angle_degree, reshape=False, cval=fill_val)

        w = WCS(self.hdu[0].header)
        try:
            # cd_matrix = w.wcs.cd
            shape = self.data.shape
            highest = min(self.data.shape)
            xs = np.random.randint(0, highest, 20)
            ys = np.random.randint(0, highest, 20)

            skys = w.pixel_to_world(xs.tolist(), ys.tolist())

            center = np.array(shape) / 2.0
            rotation_matrix = np.array([[np.cos(-angle), -np.sin(-angle)],
                                        [np.sin(-angle), np.cos(-angle)]])
            translated_coords = np.array([xs, ys]) - center[:, np.newaxis]
            new_coords = np.dot(rotation_matrix, translated_coords) + center[:, np.newaxis]
            new_coords = np.round(new_coords).astype(int)
            new_coords[0] = np.clip(new_coords[0], 0, rotated_data.shape[0] - 1)
            new_coords[1] = np.clip(new_coords[1], 0, rotated_data.shape[1] - 1)

            w = fit_wcs_from_points([new_coords[0], new_coords[1]], skys)

        except:
            self.logger.error("No WCS found in header")
            raise Unsolvable("No WCS found in header")

        temp_header = Header()
        # temp_header.extend(self.pure_header(), unique=True, update=True)
        temp_header.extend(w.to_header(), unique=True, update=True)

        if output is None:
            self.hdu[0].data = rotated_data.astype(int)
            self.hdu[0].header = temp_header
            self.hdu.flush()
            return self

        return self.from_data_header(rotated_data.astype(int), header=temp_header, output=output, override=override)

    def crop(self, x: int, y: int, width: int, height: int, output: Optional[str] = None,
             override: bool = False) -> Self:
        self.logger.info("Cropping fits")
        print(y, y + height, x, x + width)
        cropped_data = self.data[y:y + height, x:x + width]

        w = WCS(self.hdu[0].header)

        if cropped_data.size == 0:
            self.logger.error("Out of boundaries")
            raise IndexError("Out of boundaries")

        try:
            w = w[y:y + height, x:x + width]
        except:
            self.logger.error("No WCS found in header")
            raise Unsolvable("No WCS found in header")

        temp_header = Header()
        # temp_header.extend(self.pure_header(), unique=True, update=True)
        temp_header.extend(w.to_header(), unique=True, update=True)

        if output is None:
            self.hdu[0].data = cropped_data.astype(int)
            self.hdu[0].header = temp_header
            self.hdu.flush()
            return self

        return self.from_data_header(cropped_data.astype(int), header=temp_header, output=output, override=override)

    def bin(self, binning_factor: Union[int, Tuple[int, int]], func: Callable[..., float] = np.mean,
            output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Binning fits")

        if isinstance(binning_factor, int):
            binning_factor_to_use = [binning_factor] * 2
        else:
            if len(binning_factor) != 2:
                self.logger.error("Binning Factor must be a list of 2 integers")
                raise ValueError("Binning Factor must be a list of 2 integers")
            binning_factor_to_use = binning_factor
        try:
            binned_data = block_reduce(self.data, tuple(binning_factor_to_use), func=func)
            w = WCS(self.hdu[0].header)
        except ValueError:
            self.logger.error("Bad value")
            raise ValueError("Bad value")

        try:
            highest = min(self.data.shape)
            xs = np.random.randint(0, highest, 20)
            ys = np.random.randint(0, highest, 20)

            skys = w.pixel_to_world(xs.tolist(), ys.tolist())

            new_coords = [
                xs // binning_factor_to_use[0],
                ys // binning_factor_to_use[1]
            ]

            w = fit_wcs_from_points([new_coords[0], new_coords[1]], skys)
        except:
            self.logger.error("No WCS found in header")
            raise Unsolvable("No WCS found in header")

        temp_header = Header()
        # temp_header.extend(self.pure_header(), unique=True, update=True)
        temp_header.extend(w.to_header(), unique=True, update=True)

        if output is None:
            self.hdu[0].data = binned_data.astype(int)
            self.hdu[0].header = temp_header
            self.hdu.flush()
            return self

        return self.from_data_header(binned_data.astype(int), header=temp_header, output=output, override=override)

    def zero_correction(self, master_zero: Self, force: bool = False,
                        output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Zero correction")

        if "MY-ZERO" not in self.hdu[0].header or force:
            zero_corrected = subtract_bias(self.ccd, master_zero.ccd).data.astype(int)
            header = self.hdu[0].header
            header["MY-ZERO"] = master_zero.path.absolute().__str__()

            if output is None:
                self.hdu[0].data = zero_corrected
                self.hdu[0].header = header
                self.hdu.flush()
                return self

            return self.from_data_header(zero_corrected, header=header, output=output, override=override)

        self.logger.error("This Data is already zero corrected")
        raise OverCorrection("This Data is already zero corrected")

    def dark_correction(self, master_dark: Self, exposure: Optional[str] = None, force: bool = False,
                        output: Optional[str] = None, override: bool = False) -> Self:
        self.logger.info("Dark correction")

        if "MY-DARK" not in self.hdu[0].header or force:
            if exposure is None:
                options = {"dark_exposure": 1 * units.s,
                           "data_exposure": 1 * units.s}
            else:
                if exposure not in self.hdu[0].header or exposure not in master_dark.header():
                    raise CardNotFound(f"Key {exposure} not found in file, master_dark or both")

                options = {
                    "dark_exposure": float(
                        float(master_dark.header()[exposure].values[0])) * units.s,
                    "data_exposure": float(
                        float(self.hdu[0].header[exposure].values[0])) * units.s
                }

            dark_corrected = subtract_dark(
                self.ccd, master_dark.ccd,
                **options, scale=True
            )
            header = self.hdu[0].header
            header["MY-DARK"] = master_dark.path.absolute().__str__()

            if output is None:
                self.hdu[0].data = dark_corrected.data.astype(int)
                self.hdu[0].header = header
                self.hdu.flush()
                return self

            return self.__class__.from_data_header(dark_corrected.data.astype(int), header=header,
                                                   output=output, override=override)

        self.logger.error("This Data is already dark corrected")
        raise OverCorrection("This Data is already dark corrected")

    def flat_correction(self, master_flat: Self, force: bool = False, output: Optional[str] = None,
                        override: bool = False) -> Self:
        self.logger.info("Flat correction")

        if "MY-FLAT" not in self.hdu[0].header or force:
            flat_corrected = flat_correct(self.ccd, master_flat.ccd)
            header = self.hdu[0].header
            header["MY-FLAT"] = master_flat.path.absolute().__str__()

            if output is None:
                self.hdu[0].data = flat_corrected.data.astype(int)
                self.hdu[0].header = header
                self.hdu.flush()
                return self

            return self.__class__.from_data_header(flat_corrected.data.data.astype(int), header=header,
                                                   output=output, override=override)

        self.logger.error("This Data is already flat corrected")
        raise OverCorrection("This Data is already flat corrected")

    def pixels_to_skys(self, xs: Union[List[Union[int, float]], int, float],
                       ys: Union[List[Union[int, float]], int, float]) -> QTable:
        self.logger.info("Pixels to Skys")

        xs_to_use = xs if isinstance(xs, list) else [xs]
        ys_to_use = ys if isinstance(ys, list) else [ys]

        if len(xs_to_use) != len(ys_to_use):
            self.logger.error("xs and ys must be equal in length")
            raise ValueError("xs and ys must be equal in length")

        data = []

        w = WCS(self.hdu[0].header)

        for x, y in zip(xs_to_use, ys_to_use):
            sky = w.pixel_to_world(x, y)
            if not isinstance(sky, SkyCoord):
                self.logger.error("Plate is not solved")
                raise Unsolvable("Plate is not solved")

            data.append([x * units.pix, y * units.pix, sky.ra, sky.dec])

        return QTable(
            list(map(list, itertools.zip_longest(*data, fillvalue=None))),
            names=["x", "y", "ra", "dec"]
        )

    def skys_to_pixels(self, skys: SkyCoord) -> QTable:
        self.logger.info("Skys to Pixels")

        data = []

        w = WCS(self.hdu[0].header)

        for sky in skys:
            try:
                pixels = w.world_to_pixel(sky)
                data.append([sky.ra, sky.dec, float(pixels[0]) * units.pix, float(pixels[1]) * units.pix])
            except ValueError:
                self.logger.error("Plate is not solved")
                raise Unsolvable("Plate is not solved")
            if np.isnan(pixels).any():
                self.logger.error("Plate is not solved")
                raise Unsolvable("Plate is not solved")

        return QTable(
            list(map(list, itertools.zip_longest(*data, fillvalue=None))),
            names=["ra", "dec", "x", "y"]
        )

    def photometry_sep(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None) -> QTable:
        self.logger.info("Photometry sep")

        table = []

        the_header = self.header

        if exposure is None:
            exposure_to_use = 0.0
        else:
            if isinstance(exposure, (int, float)):
                exposure_to_use = exposure
            else:
                exposure_to_use = float(the_header[exposure].iloc[0])

        new_xs, new_ys = Fixer.coordinate(xs, ys)
        new_rs = Fixer.aperture(rs)
        new_headers = Fixer.header(headers)

        headers_ = []
        keys_ = []
        for new_header in new_headers:
            keys_.append(new_header)
            try:
                headers_.append(the_header[new_header])
            except KeyError:
                self.logger.warning("Couldn't find header key")
                headers_.append(None)

        data = self.data.astype(float)
        background = self.background()

        clean_d = data - background.rms()
        error = calc_total_error(
            data, background, exposure_to_use
        )
        for new_r in new_rs:
            fluxes, flux_errs, flags = sum_circle(data, new_xs, new_ys, new_r, err=error)

            for x, y, flux, flux_err, flag in zip(new_xs, new_ys, fluxes,
                                                  flux_errs, flags):
                try:
                    sky = self.pixels_to_skys(x, y).iloc[0].sky
                    ra, dec = sky.ra, sky.dec
                except Exception as e:
                    self.logger.warning(f"Could not get ra, dec. {e}")
                    ra, dec = None, None

                value = clean_d[int(x)][int(y)]
                snr = np.nan if value < 0 else math.sqrt(value)
                mag, mag_err = self.flux_to_mag(flux, flux_err,
                                                exposure_to_use)
                table.append(
                    [
                        self.file, "sep", x * units.pix, y * units.pix, ra, dec,
                                          new_r * units.pix, flux * units.adu,
                                          flux_err * units.adu, flag, snr, mag * units.mag,
                                          mag_err * units.mag, *headers_
                    ]
                )
        return QTable(
            list(map(list, itertools.zip_longest(*table, fillvalue=None))),
            names=["image", "package", "x", "y", "ra", "dec", "aperture", "flux",
                   "flux_error", "flag", "snr", "mag", "merr", *keys_]
        )

    def photometry_phu(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None) -> QTable:
        self.logger.info("Photometry photutils")

        table = []

        the_header = self.header
        data = self.data.astype(float)

        if exposure is None:
            exposure_to_use = 0.0
        else:
            if isinstance(exposure, (int, float)):
                exposure_to_use = exposure
            else:
                exposure_to_use = float(the_header[exposure].iloc[0])

        new_xs, new_ys = Fixer.coordinate(xs, ys)
        new_rs = Fixer.aperture(rs)
        new_headers = Fixer.header(headers)

        headers_ = []
        keys_ = []
        for new_header in new_headers:
            keys_.append(new_header)
            try:
                headers_.append(the_header[new_header])
            except KeyError:
                self.logger.warning("Couldn't find header key")
                headers_.append(None)

        background = self.background()

        clean_d = data - background.rms()

        for new_r in new_rs:
            apertures = CircularAperture([
                [new_x, new_y] for new_x, new_y in zip(new_xs, new_ys)
            ], r=new_r)
            error = calc_total_error(
                data, self.background(), exposure_to_use
            )
            phot_table = aperture_photometry(data, apertures, error=error)

            for phot_line in phot_table:
                value = clean_d[
                    int(phot_line["xcenter"].value)
                ][
                    int(phot_line["ycenter"].value)
                ]
                snr = np.nan if value < 0 else math.sqrt(value)
                mag, mag_err = self.flux_to_mag(
                    phot_line["aperture_sum"],
                    phot_line["aperture_sum_err"],
                    exposure_to_use
                )

                try:
                    sky = self.pixels_to_skys(phot_line["xcenter"].value, phot_line["ycenter"].value).iloc[0].sky
                    ra, dec = sky.ra, sky.dec
                except Exception as e:
                    self.logger.warning(f"Could not get ra, dec. {e}")
                    ra, dec = None, None

                table.append(
                    [
                        self.file, "phu",
                        phot_line["xcenter"].value * units.pix, phot_line["ycenter"].value * units.pix,
                        ra, dec, new_r * units.pix,
                        phot_line["aperture_sum"] * units.adu, phot_line["aperture_sum_err"] * units.adu,
                        np.nan, snr, mag * units.mag, mag_err * units.mag, *headers_]
                )

        return QTable(
            list(map(list, itertools.zip_longest(*table, fillvalue=None))),
            names=["image", "package", "x", "y", "ra", "dec", "aperture",
                   "flux", "flux_error", "flag", "snr", "mag", "merr", *keys_]
        )

    def photometry(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                   headers: Optional[Union[str, List[str]]] = None,
                   exposure: Optional[Union[str, float, int]] = None
                   ) -> QTable:
        self.logger.info("Photometry")
        table_sep = self.photometry_sep(xs, ys, rs, headers=headers, exposure=exposure)
        table_phu = self.photometry_phu(xs, ys, rs, headers=headers, exposure=exposure)
        return vstack([table_sep, table_phu])



import itertools
import math
from glob import glob
from logging import Logger, getLogger
from pathlib import Path
from random import randint
from typing import List, Union, Optional, Iterator, Literal, Callable, Tuple, Dict, Any

import numpy as np
from astropy import units
from astropy.coordinates import SkyCoord
from astropy.nddata import CCDData
from astropy.table import QTable, vstack
from astropy.units import Quantity
from astropy.visualization import ZScaleInterval
from ccdproc import Combiner
from matplotlib import pyplot as plt, animation
from sep import Background
from typing_extensions import Self

from .error import NumberOfElementError, OverCorrection
from .fits import Fits
from .dataarray_model import DataArray
from .utils import HEADER_ANY, NUMERIC, NUMERICS


def empty_phot():
    empty_table = QTable()
    empty_table['image'] = []
    empty_table['package'] = []
    empty_table['x'] = Quantity([], unit='pix')
    empty_table['y'] = Quantity([], unit='pix')
    empty_table['ra'] = Quantity([], unit='deg')
    empty_table['dec'] = Quantity([], unit='deg')
    empty_table['aperture'] = Quantity([], unit='pix')
    empty_table['flux'] = Quantity([], unit='adu')
    empty_table['flux_error'] = Quantity([], unit='adu')
    empty_table['flag'] = []
    empty_table['snr'] = []
    empty_table['mag'] = Quantity([], unit='mag')
    empty_table['merr'] = Quantity([], unit='mag')

    return empty_table


class FitsArray(DataArray):
    def __init__(self, fitses: List[Fits], logger: Optional[Logger] = None):
        if logger is None:
            self.logger = getLogger(__file__)
        else:
            self.logger = logger

        if len(fitses) == 0:
            self.logger.error("No Fits object was given")
            raise NumberOfElementError("No Fits object was given")

        self.fitses: List[Fits] = fitses

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(nof:{len(self)})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.from_paths({self.fitses.__str__()}))"

    def __len__(self) -> int:
        return len(self.fitses)

    def __iter__(self) -> Iterator[Fits]:
        for fits in self.fitses:
            yield fits

    def __getitem__(self, key: Union[int, slice]) -> Union[Fits, Self]:

        if isinstance(key, int):
            return self.fitses[key]
        elif isinstance(key, slice):
            return self.__class__(self.fitses[key])

        self.logger.error("Wrong slice")
        raise ValueError("Wrong slice")

    def __abs__(self) -> Self:
        return self.abs()

    def __add__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.add(other)

    def __radd__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.add(other)

    def __sub__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.sub(other)

    def __rsub__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.sub(other)

    def __mul__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.mul(other)

    def __rmul__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.mul(other)

    def __truediv__(self, other: Union[Self, NUMERIC]) -> Self:
        return self.div(other)

    def __pow__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        return self.pow(power)

    def __mod__(self, other: Union[Self, NUMERIC], modulo=None) -> Self:
        return self.mod(other)

    def __rmod__(self, other: Union[Self, NUMERIC], modulo=None) -> Self:
        return self.mod(other)

    def __neg__(self) -> Self:
        return self.mul(-1)

    def __pos__(self) -> Self:
        return self

    def __contains__(self, item: Fits) -> bool:
        for each in self:
            if each.is_same(item):
                return True

        return False

    def __setitem__(self, key: Union[int, slice], value: Fits) -> None:
        self.fitses[key] = value

    def __delitem__(self, key: Union[int, slice]) -> None:
        del self.fitses[key]

    @classmethod
    def from_paths(cls, paths: List[str]) -> Self:
        files = []
        for each in map(Path, paths):
            try:
                files.append(Fits(each))
            except FileNotFoundError:
                pass

        return cls(files)

    def __weight_corrector(self, weights: Optional[Union[List[str], List[Union[float, int]]]] = None):
        self.logger.info("Checking weights...")
        if weights is None:
            weights_to_use = [1] * len(self)
        else:
            weights_to_use = weights

        if isinstance(weights, List):
            the_weights = weights_to_use
        else:
            the_weights = [weights_to_use]

        if len(the_weights) != len(self):
            self.logger.error("The length of the weights is not equal to the number of fits objects")
            raise ValueError("The length of the weights is not equal to the number of fits objects")

        return the_weights

    @classmethod
    def from_pattern(cls, pattern: str) -> Self:
        return cls.from_paths(glob(pattern))

    @classmethod
    def sample(cls, n=10) -> Self:
        fitses = []
        for i in range(n):
            file = Fits.sample()
            file.shift(randint(i, 2 * i), randint(i, 2 * i)).rotate(math.radians(randint(i, 2 * i) / 10))
            fitses.append(file)

        return cls(fitses)

    def merge(self, other: Self) -> Self:
        self.logger.info(f"Merging two FitsArrays")
        self.fitses.extend(other.fitses)
        return self

    def append(self, other: Fits) -> Self:
        self.logger.info("Appending Fits to FitsArray")
        self.fitses.append(other)
        return self

    @property
    def header(self) -> QTable:
        self.logger.info("Getting fits header")
        headers = []
        for each in self:
            try:
                header = {
                    "image": each.path.absolute().__str__()
                }
                header.update(each.header)

                headers.append(header)
            except Exception as e:
                self.logger.warning(e)

        if len(headers) == 0:
            self.logger.error("Couldn't get any fits header")
            raise NumberOfElementError("Couldn't get any fits header")

        return QTable(
            [
                each
                for each in headers
            ]
        )

    @property
    def data(self) -> Iterator[np.ndarray]:
        self.logger.info("Getting fits data")
        for each in self:
            try:
                yield each.data
            except Exception as e:
                self.logger.warning(e)

    @property
    def ccd(self) -> Iterator[CCDData]:
        self.logger.info("Getting fitses CCDData")
        for each in self:
            try:
                yield each.ccd
            except Exception as e:
                self.logger.warning(e)

    def value(self, x: int, y: int) -> QTable:
        self.logger.info("Getting fitses value")
        values = []
        for each in self:
            try:
                values.append([each.path.absolute().__str__(), x, y, each.value(x, y)])
            except Exception as e:
                self.logger.warning(e)

        if len(values) == 0:
            self.logger.error("Couldn't get value for any of fits")
            raise NumberOfElementError("Couldn't get value for any of fits")

        return QTable(
            list(map(list, itertools.zip_longest(*values, fillvalue=None))),
            names=["image", "x", "y", "value"]
        )

    @property
    def files(self) -> List[str]:
        self.logger.info("Getting fits files")
        files = []
        for each in self:
            try:
                files.append(each.file)
            except Exception as e:
                self.logger.warning(e)

        if len(files) == 0:
            self.logger.error("Couldn't get files for any of fits")
            raise NumberOfElementError("Couldn't get path for any of fits")

        return files

    @property
    def stats(self) -> QTable:
        self.logger.info("Getting fits stats")
        stats = []
        for each in self:
            line = []
            try:
                line.append(each.path.absolute().__str__())
                line.extend(list(each.stats.values()))
                stats.append(line)
            except Exception as e:
                self.logger.warning(e)

        if len(stats) == 0:
            self.logger.error("Couldn't get stats for any of fits")
            raise NumberOfElementError("Couldn't get stats for any of fits")

        return QTable(
            list(map(list, itertools.zip_longest(*stats, fillvalue=None))),
            names=["image", "size", "width", "height", "min", "mean", "std", "max"]
        )

    def background(self) -> Iterator[Background]:
        self.logger.info("Getting background of fitses")
        for each in self:
            try:
                yield each.background()
            except Exception as e:
                self.logger.warning(e)

    def cosmic_cleaner(self, sigclip: float = 4.5, sigfrac: float = 0.3, objlim: float = 5.0, gain: float = 1.0,
                       readnoise: float = 6.5, satlevel: float = 65535.0, pssl: float = 0.0, niter: int = 4,
                       sepmed: bool = True, cleantype: Literal["median", "medmask", "meanmask", "idw"] = 'meanmask',
                       fsmode: Literal["median", "convolve"] = 'median',
                       psfmodel: Literal["gauss", "moffat", "gaussx", "gaussy"] = 'gauss',
                       psffwhm: float = 2.5, psfsize: int = 7, psfk: np.ndarray = None, psfbeta: float = 4.765,
                       gain_apply: bool = True) -> Self:
        self.logger.info("Cleaning fitses")
        for each in self:
            try:
                _ = each.cosmic_cleaner(sigclip=sigclip, sigfrac=sigfrac, objlim=objlim, gain=gain, readnoise=readnoise,
                                        satlevel=satlevel, pssl=pssl, niter=niter, sepmed=sepmed, cleantype=cleantype,
                                        fsmode=fsmode, psfmodel=psfmodel, psffwhm=psffwhm, psfsize=psfsize, psfk=psfk,
                                        psfbeta=psfbeta, gain_apply=gain_apply)
            except Exception as e:
                self.logger.warning(e)

            return self

    def hedit(self, keys: Union[str, List[str]], values: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None,
              delete: bool = False, value_is_key: bool = False) -> Self:
        self.logger.info("Editing the header")

        for each in self:
            _ = each.hedit(keys=keys, values=values, delete=delete, value_is_key=value_is_key)
        return self

    def hselect(self, fields: Union[str, List[str]]) -> QTable:
        self.logger.info("Selecting the header")
        if isinstance(fields, str):
            fields = [fields]

        headers = self.header
        fields_to_use = [field for field in fields if field in headers.columns]

        if len(fields_to_use) == 0:
            empty_table = QTable()
            return empty_table
        fields_to_use.insert(0, "image")
        return self.header[fields_to_use]

    def save(self, path: Union[str, Path], overwrite: bool = False) -> Self:
        self.logger.info("Saving fitses")
        fitses = []
        for each in self:
            try:
                fits = each.save(path=path, overwrite=overwrite)
                fitses.append(fits)
            except Exception as e:
                self.logger.warning(e)

        if len(fitses) == 0:
            self.logger.error("Couldn't save any fits")
            raise NumberOfElementError("Couldn't save any fits")

        return fitses

    def abs(self) -> Self:
        for each in self:
            try:
                _ = each.abs()
            except Exception as e:
                self.logger.warning(e)

        return self

    def add(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]) -> Self:
        self.logger.info("Adding value to fitses")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.add(other)
            except Exception as e:
                self.logger.warning(e)

        return self

    def sub(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]) -> Self:
        self.logger.info("Subtracting value from fitses")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.sub(other)
            except Exception as e:
                self.logger.warning(e)

        return self

    def mul(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]) -> Self:
        self.logger.info("Multiplying value with fitses")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.mul(other)
            except Exception as e:
                self.logger.error(e)

        return self

    def div(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]) -> Self:
        self.logger.info("Dividing fitses with the value")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.div(other)
            except Exception as e:
                self.logger.warning(e)

        return self

    def pow(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]) -> Self:
        self.logger.info("Rasing fitses with the value")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.pow(other)
            except Exception as e:
                self.logger.warning(e)

        return self

    def mod(self, other: Union[Self, Fits, NUMERIC, List[Union[Fits, NUMERIC]]]):
        self.logger.info("Calculating mod of fitses with the value")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.mod(other)
            except Exception as e:
                self.logger.warning(e)

        return self

    def imarith(self, other: Union[Self, NUMERIC], operator: Literal["+", "-", "*", "/", "**", "^", "%"]) -> Self:
        self.logger.info("Image arithmetic")
        if isinstance(other, List):
            if len(other) != len(self):
                self.logger.error("Number of other does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of other does not match the number of elements in FitsArray")
            others = other
        else:
            others = [other] * len(self)

        for each, other in zip(self, others):
            try:
                _ = each.imarith(other, operator)
            except Exception as e:
                self.logger.warning(e)

        return self

    def extract(self, index: int = 0, detection_sigma: float = 5.0, min_area: float = 5.0) -> QTable:
        self.logger.info("Extracting sources from fits (sep)")
        if index >= len(self):
            self.logger.error("Index out of range")
            raise IndexError("Index out of range")

        return self[index].extract(detection_sigma=detection_sigma, min_area=min_area)

    def daofind(self, index: int = 0, sigma: float = 3.0, fwhm: float = 3.0, threshold: float = 5.0) -> QTable:
        self.logger.info("Extracting sources from fits (daofind)")
        if index >= len(self):
            self.logger.error("Index out of range")
            raise IndexError("Index out of range")

        return self[index].daofind(sigma=sigma, fwhm=fwhm, threshold=threshold)

    def align(self, reference: Union[Fits, int] = 0, max_control_points: int = 50, min_area: int = 5) -> Self:
        self.logger.info("Aligning fits")
        if isinstance(reference, Fits):
            the_reference = reference
        else:
            the_reference = self[reference]

        for each in self:
            try:
                _ = each.align(the_reference, max_control_points=max_control_points, min_area=min_area)
            except Exception as e:
                self.logger.warning(e)

        return self

    def shift(self, xs: Union[List[int], int], ys: Union[List[int], int]) -> Self:
        self.logger.info("Shifting fitses")

        if type(xs) is not type(ys):
            self.logger.error("xs and ys must be same type")
            raise ValueError("xs and ys must be same type")

        if isinstance(xs, List):
            if len(xs) != len(ys) != len(self):
                self.logger.error("Number of coordinates does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of coordinates does not match the number of elements in FitsArray")
            new_xs = xs
            new_ys = ys
        else:
            new_xs = [xs] * len(self)
            new_ys = [ys] * len(self)

        for each, x, y in zip(self, new_xs, new_ys):
            try:
                _ = each.shift(x, y)
            except Exception as e:
                self.logger.warning(e)

        return self

    def rotate(self, angles: Union[List[NUMERIC], NUMERIC]) -> Self:
        self.logger.info("Rotating fitses")
        if isinstance(angles, List):
            if len(angles) != len(self):
                self.logger.error("Number of angles does not match the number of elements in FitsArray")
                raise NumberOfElementError("Number of angles does not match the number of elements in FitsArray")
            new_angles = angles
        else:
            new_angles = [angles] * len(self)

        for each, angle in zip(self, new_angles):
            try:
                _ = each.rotate(angle)
            except Exception as e:
                self.logger.warning(e)

        return self

    def crop(self, xs: Union[int, List[int]], ys: Union[int, List[int]],
             widths: Union[int, List[int]], heights: Union[int, List[int]]) -> Self:
        self.logger.info("Cropping fitses")
        if type(xs) is not type(ys) is not type(widths) is not type(heights):
            self.logger.error("xs, ys, widths and heights must be same type")
            raise ValueError("xs, ys, widths and heights must be same type")

        if isinstance(xs, List):
            if len(xs) != len(ys) != len(widths) != len(heights) != len(self):
                self.logger.error("Number of coordinates and dimensions does not match the "
                                  "number of elements in FitsArray")
                raise NumberOfElementError("Number of coordinates and dimensions does not match the "
                                           "number of elements in FitsArray")
            new_xs = xs
            new_ys = ys
            new_widths = widths
            new_heights = heights
        else:
            new_xs = [xs] * len(self)
            new_ys = [ys] * len(self)
            new_widths = [widths] * len(self)
            new_heights = [heights] * len(self)

        for each, x, y, w, h in zip(self, new_xs, new_ys, new_widths, new_heights):
            try:
                _ = each.crop(x, y, w, h)
            except Exception as e:
                self.logger.warning(e)

        return self

    def bin(self, binning_factors: Union[Union[int, Tuple[int, int]], List[Union[int, Tuple[int, int]]]],
            func: Callable[..., float] = np.mean) -> Self:
        self.logger.info("Binning fitses")

        if isinstance(binning_factors, List):
            if len(binning_factors) != len(self):
                self.logger.error("Number of binning_factors number of elements in FitsArray")
                raise NumberOfElementError("Number of binning_factors number of elements in FitsArray")
            new_binning_factors = binning_factors
        else:
            new_binning_factors = [binning_factors] * len(self)

        for each, binning_factor in zip(self, new_binning_factors):
            try:
                _ = each.bin(binning_factor)
            except Exception as e:
                self.logger.warning(e)

        return self

    def show(self, points: QTable = None, scale: bool = True, interval: float = 1.0) -> None:
        self.logger.info("Animation fitses")
        fig = plt.figure()

        if scale:
            zscale = ZScaleInterval()
        else:
            def zscale(x):
                return x

        im = plt.imshow(zscale(self[0].data), cmap="Greys_r", animated=True)
        plt.xticks([])
        plt.yticks([])

        def updatefig(args):
            im.set_array(zscale(self[args % len(self)].data))
            return im,

        _ = animation.FuncAnimation(
            fig, updatefig, interval=interval, blit=True,
            cache_frame_data=False
        )
        plt.show()

    def solve_field(self, api_key: str, reference: Union[Fits, int] = 0, solve_timeout: int = 120,
                    force_image_upload: bool = False, max_control_points: int = 50, min_area: int = 5,
                    soft_solve: bool = True) -> Self:
        self.logger.info("Solving plate")

        return self

    def group_by(self, groups: Union[str, List[str]]) -> Dict[Any, Self]:
        self.logger.info("Grouping fitses")
        if isinstance(groups, List):
            the_groups = groups
        else:
            the_groups = [groups]

        if len(the_groups) < 1:
            return dict()

        headers = self.header.to_pandas()

        for group in the_groups:
            if group not in headers.columns:
                headers[group] = np.nan

        grouped = {}
        for keys, df in headers.groupby(groups, dropna=False):
            grouped[keys] = self.__class__.from_paths(df["image"].tolist())

        return grouped

    def combine(self, method: Literal["average", "mean", "median", "sum"] = "average",
                clipping: Literal[None, "sigma", "minmax"] = None,
                weights: Optional[List[Union[float, int]]] = None) -> Fits:
        self.logger.info("Combining fitses")
        if method not in ["average", "mean", "median", "sum"]:
            self.logger.error("Method must be one of average, mean, median, sum")
            raise ValueError("Method must be one of average, mean, median, sum")

        if clipping not in [None, "sigma", "minmax"]:
            self.logger.error("Clipping must be one of None, sigma, minmax")
            raise ValueError("Clipping must be one of None, sigma, minmax")

        combiner = Combiner(self.ccd)

        if weights is None:
            the_weights = [1] * len(self)
        else:
            the_weights = weights

        if len(the_weights) != len(self):
            self.logger.error("Length of weights must be equal to number of Fits")
            raise ValueError("Length of weights must be equal to number of Fits")

        if clipping is not None:
            if "sigma".startswith(clipping):
                combiner.sigma_clipping()
            else:
                combiner.minmax_clipping()

        combiner.weights = np.array(the_weights)

        if "median".startswith(method.lower()):
            the_data = combiner.median_combine().data
        elif "sum".startswith(method.lower()):
            the_data = combiner.sum_combine().data
        else:
            the_data = combiner.average_combine().data

        return Fits.from_data_header(data=the_data)

    def zero_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None) -> Fits:
        self.logger.info("Zero combining fitses")
        return self.combine(method=method, clipping=clipping)

    def dark_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None,
                     weights: Optional[Union[List[str], List[Union[float, int]]]] = None) -> Fits:
        self.logger.error("Dark combining fitses")

        the_weights = self.__weight_corrector(weights)
        return self.combine(method=method, clipping=clipping, weights=the_weights)

    def flat_combine(self, method: Literal["average", "mean", "median"] = "median",
                     clipping: Literal[None, "sigma", "minmax"] = None,
                     weights: Optional[Union[List[str], List[Union[float, int]]]] = None) -> Fits:
        self.logger.error("Flat combining fitses")

        the_weights = self.__weight_corrector(weights)
        return self.combine(method=method, clipping=clipping, weights=the_weights)

    def zero_correction(self, master_zero: Self, force: bool = False) -> Self:
        self.logger.info("Zero correction")
        fits_array = []
        for fits in self:
            try:
                zero_corrected = fits.zero_correction(master_zero, force=force)
                fits_array.append(zero_corrected)
            except OverCorrection:
                fits_array.append(fits)
            except Exception as error:
                self.logger.warning(error)

        return self.__class__(fits_array)

    def dark_correction(self, master_dark: Self, exposure: Optional[str] = None, force: bool = False) -> Self:
        self.logger.info("Dark correction")
        fits_array = []
        for fits in self:
            try:
                dark_corrected = fits.dark_correction(master_dark, exposure=exposure, force=force)
                fits_array.append(dark_corrected)
            except OverCorrection:
                fits_array.append(fits)
            except Exception as error:
                self.logger.warning(error)

        return self.__class__(fits_array)

    def flat_correction(self, master_flat: Self, exposure: Optional[str] = None, force: bool = False) -> Self:
        self.logger.info("Flat correction")
        fits_array = []
        for fits in self:
            try:
                dark_corrected = fits.flat_correction(master_flat, force=force)
                fits_array.append(dark_corrected)
            except OverCorrection:
                fits_array.append(fits)
            except Exception as error:
                self.logger.warning(error)

        return self.__class__(fits_array)

    def photometry_sep(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None) -> QTable:
        self.logger.info("Photometry sep")
        photometry = []
        for fits in self:
            try:
                phot = fits.photometry_sep(xs, ys, rs, headers=headers, exposure=exposure)
                photometry.append(phot)
            except NumberOfElementError:
                self.logger.error("The length of Xs and Ys must be equal")
                raise NumberOfElementError("The length of Xs and Ys must be equal")
            except Exception as error:
                self.logger.warning(error)

        if len(photometry) == 0:
            self.logger.warning("Couldn't do photometry")
            raise NumberOfElementError("Couldn't do photometry")

        return vstack(photometry)

    def photometry_phu(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None) -> QTable:
        self.logger.info("Photometry photutils")
        photometry = []
        for fits in self:
            try:
                phot = fits.photometry_phu(xs, ys, rs, headers=headers, exposure=exposure)
                photometry.append(phot)
            except NumberOfElementError:
                self.logger.error("The length of Xs and Ys must be equal")
                raise NumberOfElementError("The length of Xs and Ys must be equal")
            except Exception as error:
                self.logger.warning(error)

        if len(photometry) < 1:
            self.logger.warning("Couldn't do photometry")
            raise NumberOfElementError("Couldn't do photometry")

        return vstack(photometry)

    def photometry(self, xs: NUMERICS, ys: NUMERICS, rs: NUMERICS,
                   headers: Optional[Union[str, List[str]]] = None,
                   exposure: Optional[Union[str, float, int]] = None
                   ) -> QTable:
        self.logger.info("Photometry")
        photometry = []
        for fits in self:
            try:
                phot = fits.photometry(xs, ys, rs, headers=headers, exposure=exposure)
                photometry.append(phot)
            except NumberOfElementError:
                self.logger.error("The length of Xs and Ys must be equal")
                raise NumberOfElementError("The length of Xs and Ys must be equal")
            except Exception as error:
                self.logger.warning(error)

        if len(photometry) == 0:
            self.logger.warning("Couldn't do photometry")
            raise NumberOfElementError("Couldn't do photometry")

        return vstack(photometry)

    def pixels_to_skys(self, xs: Union[List[Union[int, float]], int, float],
                       ys: Union[List[Union[int, float]], int, float]) -> QTable:
        self.logger.info("Pixels to Skys")
        data = []
        for each in self:
            try:
                values = each.pixels_to_skys(xs, ys)
                coordinates = values.to_pandas().to_numpy().tolist()
                for i in range(len(coordinates)):
                    coordinates[i].insert(0, each.path.absolute().__str__())
                data.extend(coordinates)
            except Exception as e:
                self.logger.warning(e)

        if len(data) == 0:
            self.logger.error("Couldn't find pixels to skys")
            raise NumberOfElementError("Couldn't find pixels to skys")

        table = QTable(
            list(map(list, itertools.zip_longest(*data, fillvalue=None))),
            names=["image", "x", "y", "ra", "dec"],
        )
        table["x"] *= units.pix
        table["y"] *= units.pix
        table["ra"] *= units.deg
        table["dec"] *= units.deg
        return table

    def skys_to_pixels(self, skys: SkyCoord) -> QTable:
        data = []
        for each in self:
            try:
                values = each.skys_to_pixels(skys)
                coordinates = values.to_pandas().to_numpy().tolist()
                for i in range(len(coordinates)):
                    coordinates[i].insert(0, each.path.absolute().__str__())
                data.extend(coordinates)
            except Exception as e:
                self.logger.warning(e)

        if len(data) == 0:
            self.logger.error("Couldn't find pixels to skys")
            raise NumberOfElementError("Couldn't find pixels to skys")

        table = QTable(
            list(map(list, itertools.zip_longest(*data, fillvalue=None))),
            names=["image", "ra", "dec", "x", "y"],
        )
        table["x"] *= units.pix
        table["y"] *= units.pix
        table["ra"] *= units.deg
        table["dec"] *= units.deg
        return table



from abc import ABC, abstractmethod
from typing import Dict, Any, Union, List, Literal, Tuple, Optional, Callable
from typing_extensions import Self
from pathlib import Path
from matplotlib import pyplot as plt
from sep import Background
import numpy as np
from astropy.table import QTable
from astropy.nddata import CCDData
from astropy.coordinates import SkyCoord

from myraflib.utils import NUMERIC, NUMERICS, HEADER_ANY


class Data(ABC):

    @abstractmethod
    def __str__(self) -> str:
        """String representation of Data"""

    @abstractmethod
    def __repr__(self) -> str:
        """Representation of Data"""

    @abstractmethod
    def __del__(self) -> None:
        """Destructor of data"""

    @abstractmethod
    def __abs__(self) -> Self:
        """Absolute value of data"""

    @abstractmethod
    def __add__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (+)"""

    @abstractmethod
    def __iadd__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (+=)"""

    @abstractmethod
    def __sub__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (-)"""

    @abstractmethod
    def __isub__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (-=)"""

    @abstractmethod
    def __mul__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (*)"""

    @abstractmethod
    def __imul__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (*=)"""

    @abstractmethod
    def __truediv__(self, other: Union[Self, NUMERIC]) -> Self:
        """Operator overloading (/)"""

    @abstractmethod
    def __idiv__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (/=)"""

    @abstractmethod
    def __pow__(self, power: Union[Self, NUMERIC], modulo=None) -> Self:
        """Operator overloading (**)"""

    @abstractmethod
    def __ipow__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (**=)"""

    @abstractmethod
    def __mod__(self, other: Union[Self, NUMERIC], modulo=None) -> Self:
        """Operator overloading (%)"""

    @abstractmethod
    def __imod__(self, other: Union[Self, NUMERIC]) -> Self:
        """Inplace operator overloading (%=)"""

    @abstractmethod
    def __eq__(self, other: Self) -> bool:
        """Operator overloading (==)"""

    @abstractmethod
    def __ne__(self, other: Self) -> Self:
        """Operator overloading (!=)"""

    @abstractmethod
    def __neg__(self) -> Self:
        """Negative operator overloading"""

    @abstractmethod
    def __pos__(self) -> Self:
        """Positive operator overloading"""

    @property
    @abstractmethod
    def file(self) -> str:
        """Returns the fits file's path"""

    @classmethod
    @abstractmethod
    def from_path(cls, path: str) -> Self:
        """Creates a Data Object from a given path"""

    @classmethod
    @abstractmethod
    def sample(cls) -> Self:
        """Creates a sample Data"""

    @abstractmethod
    def is_same(self, other: Self) -> bool:
        """Checks if data is identical to other"""

    @property
    @abstractmethod
    def ccd(self) -> CCDData:
        """Returns the CCDData of fits"""

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        """Returns the data of the given fits file"""

    @abstractmethod
    def value(self, x: int, y: int) -> NUMERIC:
        """Returns value of a given coordinates"""

    @property
    @abstractmethod
    def header(self) -> Dict[str, Any]:
        """Returns the header of the given fits file"""

    @property
    @abstractmethod
    def stats(self) -> Dict[str, Any]:
        """Returns the stats of the given fits file"""

    @abstractmethod
    def background(self) -> Background:
        """Returns the background of the given fits file"""

    @abstractmethod
    def cosmic_cleaner(self, sigclip: float = 4.5, sigfrac: float = 0.3, objlim: float = 5.0, gain: float = 1.0,
                       readnoise: float = 6.5, satlevel: float = 65535.0, pssl: float = 0.0, niter: int = 4,
                       sepmed: bool = True, cleantype: Literal["median", "medmask", "meanmask", "idw"] = 'meanmask',
                       fsmode: Literal["median", "convolve"] = 'median',
                       psfmodel: Literal["gauss", "moffat", "gaussx", "gaussy"] = 'gauss',
                       psffwhm: float = 2.5, psfsize: int = 7, psfk: np.ndarray = None, psfbeta: float = 4.765,
                       gain_apply: bool = True, output: Optional[str] = None, override: bool = False) -> Self:
        """Cleans the fits file from cosmic rays"""

    @abstractmethod
    def hedit(self, keys: Union[str, List[str]], values: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None,
              comment: Optional[Union[HEADER_ANY, List[HEADER_ANY]]] = None, delete: bool = False,
              value_is_key: bool = False, output: Optional[str] = None, override: bool = False) -> Self:
        """Edits header of the given file"""

    @abstractmethod
    def save(self, path: Union[str, Path], overwrite: bool = False) -> Self:
        """Saves the fits file to the given path"""

    @abstractmethod
    def abs(self, output: Optional[str] = None, override: bool = False) -> Self:
        """Return absolute value of DataArray"""

    @abstractmethod
    def add(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        """Adds either another numeric or another fits to this fits file"""

    @abstractmethod
    def sub(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        """Subtracts either another numeric or another fits from this fits file"""

    @abstractmethod
    def mul(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        """Multiplies either another numeric or another fits from this fits file"""

    @abstractmethod
    def div(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        """Divides either another numeric or another fits to this fits file"""

    def pow(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False) -> Self:
        """Raises the data to given value"""

    def mod(self, other: Union[Self, NUMERIC], output: Optional[str] = None, override: bool = False):
        """Modular arithmetic operation"""

    @abstractmethod
    def imarith(self, other: Union[Self, NUMERIC], operator: Literal["+", "-", "*", "/", "**", "^", "%"],
                output: Optional[str] = None, override: bool = False) -> Self:
        """Makes arithmetic operation to fits"""

    @abstractmethod
    def extract(self, detection_sigma: float = 5.0, min_area: float = 5.0) -> QTable:
        """Extracts sources from fits"""

    @abstractmethod
    def daofind(self, sigma: float = 3.0, fwhm: float = 3.0, threshold: float = 5.0) -> QTable:
        """Extracts sources from fits using daofind"""

    @abstractmethod
    def align(self, reference: Self, max_control_points: int = 50, detection_sigma: float = 5.0, min_area: int = 5,
              output: Optional[str] = None, override: bool = False) -> Self:
        """Aligns this with given fits"""

    @abstractmethod
    def show(self, ax: plt.Axes = None, sources: QTable = None, scale: bool = True) -> None:
        """Shows fits data"""

    @abstractmethod
    def solve_field(self, api_key: str) -> Self:
        """WCS solves"""

    @abstractmethod
    def coordinate_picker(self, scale: bool = True) -> QTable:
        """Shows the Image using matplotlib and returns a list of coordinates picked by user"""

    @abstractmethod
    def shift(self, x: NUMERIC, y: NUMERIC, output: Optional[str] = None, override: bool = False) -> Self:
        """Shifts the image"""

    @abstractmethod
    def rotate(self, angle: NUMERIC, output: Optional[str] = None, override: bool = False) -> Self:
        """Rotates the image"""

    @abstractmethod
    def crop(self, x: int, y: int, width: int, height: int, output: Optional[str] = None,
             override: bool = False) -> Self:
        """Crop the data of `Fits` object"""

    @abstractmethod
    def bin(self, binning_factor: Union[int, Tuple[int, int]], func: Callable[..., float] = np.mean,
            output: Optional[str] = None, override: bool = False) -> Self:
        """Bin the data of `Fits` object"""

    @abstractmethod
    def zero_correction(self, master_zero: Self, force: bool = False,
                        output: Optional[str] = None, override: bool = False) -> Self:
        """Zero corrects the fits"""

    @abstractmethod
    def dark_correction(self, master_dark: Self, exposure: Optional[str] = None, force: bool = False,
                        output: Optional[str] = None, override: bool = False) -> Self:
        """Dark corrects the fits"""

    @abstractmethod
    def flat_correction(self, master_flat: Self, force: bool = False, output: Optional[str] = None,
                        override: bool = False) -> Self:
        """Flat corrects the fits"""

    @abstractmethod
    def pixels_to_skys(self, xs: Union[List[Union[int, float]], int, float],
                       ys: Union[List[Union[int, float]], int, float]) -> QTable:
        """Calculate Sky Coordinate of given Pixel"""

    @abstractmethod
    def skys_to_pixels(self, skys: SkyCoord) -> QTable:
        """Calculate Pixel Coordinate of given Sky"""

    @abstractmethod
    def photometry_sep(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None
                       ) -> QTable:
        """Does a photometry using sep"""

    @abstractmethod
    def photometry_phu(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                       headers: Optional[Union[str, List[str]]] = None,
                       exposure: Optional[Union[str, float, int]] = None
                       ) -> QTable:
        """Does a photometry using photutils"""

    @abstractmethod
    def photometry(self, xs: Union[NUMERIC, NUMERICS], ys: Union[NUMERIC, NUMERICS], rs: Union[NUMERIC, NUMERICS],
                   headers: Optional[Union[str, List[str]]] = None,
                   exposure: Optional[Union[str, float, int]] = None
                   ) -> QTable:
        """Does a photometry using both sep and photutils"""
